from collections import namedtuple, defaultdict
from .message import Message

class FlightCollection:
	"""A collection of :py:class:`FlightCollectionEntry`'s, stored by hexident.

	It does not provide groundbreaking new features, but it's a helper that stores flights for easy querying.
	An instance can be queried (``collection['xxxxx']``) and iterated through.
	"""

	def __init__(self):
		self._dictionary = defaultdict(FlightCollectionEntry)

	def __len__(self):
		return len(self._dictionary)

	def __getitem__(self, name):
		if name in self._dictionary:
			return self._dictionary[name]
		else:
			raise KeyError(name)

	def __iter__(self):
		return iter(self._dictionary.values())

	def add(self, message):
		"""Adds a message to this collection.

		Args:
			message (:py:class:`Message`): message to add

		"""
		if not isinstance(message, Message):
			message = Message.from_string(message)
		self._dictionary[message.hexident].append(message)

	def flights(self):
		"""All stored flights.

		Returns:
			list: List of :py:class:`py1090.collection.FlightCollectionEntry`

		"""
		return self._dictionary.values()

	def add_list(self, iterable):
		"""Adds multiple messages to the collection.

		Args:
			iterable (iterable): List, file-like object or :py:class:`Connection` which contains all lines to be added.
				Calls :py:meth:`Message.from_string` on each item if it is not a :py:class:`Message` already.

		"""
		for message in iterable:
			if not isinstance(message, Message):
				message = Message.from_string(message)
			self.add(message)

class FlightCollectionEntry:
	"""Entry of a :py:class:`FlightCollection`. Allows for easy querying flight data, since one single message does not contain all the data
	information about a flight.

	See Also:
		:py:class:`Message`
			For details about the accuracy of the positional information provided by this class.

	"""
	def __init__(self):
		self.messages = []
		self.hexident = None

	def append(self, message):
		"""Adds a message that should belong to this collection.

		Will raise an exception if the message hexident does not match the one of the flight. Only adds message of type 'MSG'
		(transmission messages).

		Args:
			message (Message): the message to append

		"""
		if self.hexident is None:
			self.hexident = message.hexident
		else:
			if self.hexident != message.hexident:
				raise ValueError("Added message of different hexident.")

		if message.message_type == 'MSG':
			self.messages.append(message)

	@property
	def last_position(self):
		"""Finds the last known position of the flight (by iterating backwards through collected messages).

		Returns:
			tuple: a tuple of :py:class:`float` if the position was ever recorded, (None, None) otherwise.

		"""
		for message in reversed(self.messages):
			if message.latitude and message.longitude:
				return (message.latitude, message.longitude)
		return None, None

	@property
	def last_altitude(self):
		"""Finds the last known altitude of the flight (by iterating backwards through collected messages).

		Returns:
			float: the altitude in feet, None otherwise.

			"""
		for message in reversed(self.messages):
			if message.altitude:
				return message.altitude
		return None

	@property
	def path(self):
		"""Reconstructs the flight path. Yields it as an iterator.

		Yields:
			tuple: (lat, lon, alt) describing the latitude, longitude and altitude of a message

		"""
		for message in self.messages:
			lat, lon, alt = message.latitude, message.longitude, message.altitude
			if lat and lon and alt:
				yield (lat, lon, alt)


	def __iter__(self):
		return iter(self.messages)




from enum import Enum
from datetime import datetime

# http://www.homepages.mcb.net/bones/SBS/Article/Barebones42_Socket_Data.htm
def _parse_datetime(datestr, timestr):
	date = datetime.strptime(datestr, '%Y/%m/%d')

	timestr, separator, millisstr = timestr.rpartition('.')
	time = datetime.strptime(timestr, '%H:%M:%S')

	seconds = float(separator + millisstr)
	time = time.replace(microsecond=int(seconds*1000000))

	return datetime.combine(date.date(), time.time())

def _parse_bool(str):
	if str.lower() in ('true', 'y', 'yes', 'on', '1'):
		return True
	elif str.lower() in ('false', 'n', 'no', 'off', '0'):
		return False

def _dump_datetime(time):
	str = datetime.strftime(time, '%Y/%m/%d,%H:%M:%S')
	str += '.{:03d}'.format(int(time.microsecond/1000))
	return str

def _dump_bool(value):
	if value == True:
		return '1'
	else:
		return '0'

def _dump_or_none(value, func=str):
	if value is None:
		return ''
	else:
		return func(value)

class Message:
	"""Abstract representation of the information contained in a BaseStation line.

	The fields are listed in the order of appearance in the format. A sample line looks like this: ::

		MSG,3,111,11111,3C49CC,111111,2015/05/01,17:06:55.370,2015/05/01,17:06:55.326,,24400,,,50.65931,6.67709,,,,,,0

	This message contains the coordinates (50.65931, 6.67709) and altitude of 24400ft of an aircraft with the ADS-B hexident 3C49CC.
	The third, fourth and sixth fields have been set to '111...' which indicates that the demodulator did not support these fields.

	The attributes of this object will be set according to their values or left at None if no information is available.

	To create an instance of this object, use :py:meth:`Message.from_string`. To serialize back into a BaseStation string, use
	:py:meth:`Message.to_string`.

	See Also:
		`John Woodsides	SBS Station Tutorial <http://avphotosonline.org.uk/Bones/SBS/Article/Barebones42_Socket_Data.htm>`_
			Source of most of the information provided about the socket format here.

	Attributes:
		message_type (str): The message type: MSG, STA, ID, AIR, SEL or CLK. Usually only messages of type MSG (transmission messages)
			are of interest. Most softwares (e.g. :program:`dump1090`) do not even support any other message types.

			====    =========================    ==========================================================================================
			ID      Type                         Description
			====    =========================    ==========================================================================================
			SEL     SELECTION CHANGE MESSAGE     Generated when the user changes the selected aircraft in BaseStation.
			ID      NEW ID MESSAGE               Generated when an aircraft being tracked sets or changes its callsign.
			AIR     NEW AIRCRAFT MESSAGE         Generated when the SBS picks up a signal for an aircraft that it isn't currently tracking.
			STA     STATUS CHANGE MESSAGE        Generated when an aircraft's status changes according to the timeout values in the Data
			                                     Settings menu.
			CLK     CLICK MESSAGE                Generated when the user doubleclicks (or presses return) on an aircraft (i.e. to bring
			                                     up the aircraft details window).
			MSG     TRANSMISSION MESSAGE         Generated by the aircraft. There are eight different MSG types.
			====    =========================    ==========================================================================================


		transmission_type (int): Message subtype 1-8. Only used for transmission messages (MSG):

			=====    ============    ==========================================================================================
			Value    Spec            Description
			=====    ============    ==========================================================================================
			1        DF17 BDS 0,8    ES Identification and Category
			2        DF17 BDS 0,6    ES Surface Position Message (Triggered by nose gear squat switch.)
			3        DF17 BDS 0,5    ES Airborne Position Message
			4        DF17 BDS 0,9    ES Airborne Velocity Message
			5        DF4, DF20       Surveillance Alt Message (Triggered by ground radar. Not CRC secured.)
			6        DF5, DF21       Surveillance ID Message (Triggered by ground radar. Not CRC secured.)
			7        DF16            Air To Air Message (Triggered from TCAS)
			8        DF11            All Call Reply (Broadcast but also triggered by ground radar)
			=====    ============    ==========================================================================================

			Transmission messages of type 5 and 6 will only be output if the aircraft has previously sent a MSG 1, 2, 3, 4 or 8 signal.

			Abbr.: DF = Downlink Format, BDS = Binary Data Store, TCAS = Traffic Alert and Collision Avoidance System

		session_id (int): Database Session record number.

		aircraft_id (int): Database Aircraft record number.

		hexident (str): Aircraft Mode S hexadecimal code, given in upper case.

		flight_id (str): Database Flight record number, given in upper case.

		generation_time (datetime.datetime): Generation time of the message.

		record_time (datetime.datetime): Recording time of the message.

		callsign (str): An eight digit flight ID - can be flight number or registration (or even nothing).

		altitude (int): Mode C altitude. Height relative to 1013.2mb (Flight Level). Not height above MSL (mean sea level).

		ground_speed (float): Speed over ground (not indicated airspeed) in knots (use `helpers.knots_to_kmh` or `helpers.knots_to_mps` to convert).

		track (float): Track of aircraft (not heading), in degrees. Derived from the velocity E/W and velocity N/S.

		latitude (float): Latitude in degrees, North and East positive. Airborne valid for 4 decimal places (about 5.1m), on ground valid for 5
			places (about 1.25m). Since not all positions can be encoded, some information has to be added by the client. This is not implemented in
			py1090 yet.

		longitude (float): Longitude in degrees, South and West negative. For accuracy information see `latitude`.

		vertical_rate (int): Vertical velocity in ft/s, 64ft/s resolution.

		squawk (str): Assigned Mode A squawk code.

		squawk_alert (bool): Flag to indicate squawk has changed.

		emergency (bool): Flag to indicate emergency code has been set.

		spi (bool): Special Purpose identification. Flag to indicate transponder ident has been activated.

		on_ground (bool): Flag to indicate ground squat switch is active.
	"""

	def __init__(self):
		self.message_type = None
		self.transmission_type = None
		self.session_id = None
		self.aircraft_id = None
		self.hexident = None
		self.flight_id = None
		self.generation_time = None
		self.record_time = None
		self.callsign = None
		self.altitude = None
		self.ground_speed = None
		self.track = None
		self.latitude = None
		self.longitude = None
		self.vertical_rate = None
		self.squawk = None
		self.squawk_alert = None
		self.emergency = None
		self.spi = None
		self.on_ground = None


	def parse_string(self, string):
		parts = string.strip().split(',')

		if parts[0]:
			self.message_type = parts[0].upper()

		if parts[1]:
			self.transmission_type = int(parts[1])

		if parts[2] and parts[2] != '111':
			self.session_id = int(parts[2])

		if parts[3] and parts[3] != '11111':
			self.aircraft_id = int(parts[3])

		if parts[4]:
			self.hexident = parts[4].upper()

		if parts[5] and parts[5] != '111111':
			self.flight_id = parts[5].upper()

		if len(parts[6]) > 0 and len(parts[7]) > 0:
			self.generation_time = _parse_datetime(parts[6], parts[7])

		if len(parts[8]) > 0 and len(parts[8]) > 0:
			self.record_time = _parse_datetime(parts[8], parts[9])

		if len(parts) > 10 and parts[10]:
			self.callsign = parts[10].upper()

		if self.message_type == 'MSG':
			if parts[11]:
				self.altitude = int(parts[11])

			# This is a workaround for a bug for rtl1090 output.
			# For subtype 7, only 21 fields are sent, and only altitude and on_ground is set.
			if self.transmission_type == 7 and len(parts) == 21:
				if parts[-1]:
					self.on_ground = _parse_bool(parts[-1])
				return

			if parts[12]:
				self.ground_speed = float(parts[12])

			if parts[13]:
				self.track = float(parts[13])

			if parts[14]:
				self.latitude = float(parts[14])

			if parts[15]:
				self.longitude = float(parts[15])

			if parts[16]:
				self.vertical_rate = int(parts[16])

			if parts[17]:
				self.squawk = int(parts[17])

			if parts[18]:
				self.squawk_alert = _parse_bool(parts[18])

			if parts[19]:
				self.emergency = _parse_bool(parts[19])

			if parts[20]:
				self.spi = _parse_bool(parts[20])

			if parts[21]:
				self.on_ground = _parse_bool(parts[21])


	def to_string(self):
		"""Serializes the message into the BaseStation format.

		:rtype: str
		"""
		format_coordinates = lambda x: '{:.5f}'.format(x)

		return ','.join((
			_dump_or_none(self.message_type),
			_dump_or_none(self.transmission_type),
			_dump_or_none(self.session_id),
			_dump_or_none(self.aircraft_id),
			_dump_or_none(self.hexident),
			_dump_or_none(self.flight_id),
			_dump_or_none(self.generation_time, _dump_datetime),
			_dump_or_none(self.record_time, _dump_datetime),
			_dump_or_none(self.callsign),
			_dump_or_none(self.altitude),
			_dump_or_none(self.ground_speed),
			_dump_or_none(self.track),
			_dump_or_none(self.latitude, format_coordinates),
			_dump_or_none(self.longitude, format_coordinates),
			_dump_or_none(self.vertical_rate),
			_dump_or_none(self.squawk),
			_dump_or_none(self.squawk_alert, _dump_bool),
			_dump_or_none(self.emergency, _dump_bool),
			_dump_or_none(self.spi, _dump_bool),
			_dump_or_none(self.on_ground, _dump_bool),
		)) + '\n'

	@classmethod
	def from_string(cls, string):
		"""Creates a new message from a BaseStation format line.

		:param str string: String to parse.
		:rtype: `Message`
		"""
		message = cls()
		message.parse_string(string)
		return message

	@staticmethod
	def iter_messages(iterator):
		"""Iterates an through an iterator and yields one message per line ::

			for message in Message.iter_messages(file):
				print(message)

		:param iterable iterator:
		"""
		for item in iterator:
			yield Message.from_string(item)



import socket
import select
import io

from .message import Message

class Connection(io.TextIOWrapper):
	"""File like object which can be used to read BaseStation messages from a TCP server.

	The connection can be used as contextmanager (:keyword:`with`-block) and as iterator: ::

		with Connection() as connection:
			for line in connection:
				print(line)

	Args:
		host(str): IP or hostname
		port(int): Port number

	"""

	def __init__(self, host="localhost", port=30003):
		self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self.socket.connect((host, port))

		# have to keep a reference of the wrapper, otherwise it is garbage collected at the end of
		# the __init__ function and we will get an error 'I/O operation on closed file'.
		self._wrapper = self.socket.makefile('r')
		super().__init__(self._wrapper.buffer, encoding=self._wrapper.encoding, errors=self._wrapper.errors)

	def __next__(self):
		# make this iterator non-blocking for Ctrl+C
		while not self.has_data():
			pass
		return super().__next__()


	def has_data(self):
		"""Checks if the socket currently has data available for reading (using :py:func:`select.select`).

		Returns:
			bool: True if there is data available for reading, False otherwise.

		"""
		rlist, _, _ = select.select([ self.socket ], [], [], 0)
		return len(rlist) > 0

	def readmessage(self):
		"""Reads a single line from the connection, parses it via :py:meth:`Message.from_string` and returns it.

		Returns:
			Message: next unread message from socket
		"""
		return Message.from_string(self.readline())

	def __del__(self):
		self.socket.close()



"""
This module provides a pythonic way to process ADS-B messages. It can work with any software that provides BaseStation-like output,
for example `dump1090 <https://github.com/MalcolmRobb/dump1090>`_.
"""
from .connection import Connection
from .message import Message
from .collection import FlightCollection


#!/usr/bin/env python

from distutils.core import setup

setup(name='py1090',
      version='1.1',
      description='ADS-B messages from BaseStation handled in a pythonic way.',
      author='Jonas Lieb',
      author_email='jonas@jonaslieb.com',
      maintainer='kwadra',
      maintainer_email='py1090@demosoft.org',
      url='http://www.jonaslieb.com/',
      packages=['py1090'],
      requires=[],
      license='Apache License, Version 2.0',
      classifiers=[
      	'Development Status :: 3 - Alpha',
      	'Intended Audience :: Education',
      	'Intended Audience :: Science/Research',
      	'License :: OSI Approved :: Apache Software License',
      	'Programming Language :: Python :: 3',
      	'Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator',
      ],
      python_requires='>=3.6',
	)



import math

EARTH_RADIUS = 6371008.7714 # m
r"""The average earth radius :math:`R_0`. It is defined as the mean radius of the semi-axes.
The values are taken from the WGS 84 (World Geodetic System 1984) ellipsoid
`(definition of the Department Of Defense, Jan. 2000, p. 37) <http://earth-info.nga.mil/GandG/publications/tr8350.2/wgs84fin.pdf>`_ .

.. math::

	R_0 = 6371008.7714 \mathrm{m}

"""

def distance_between(lat1, lon1, lat2, lon2):
	r"""Calculates the distance between two locations, in meters, using the `Haversine <http://en.wikipedia.org/wiki/Haversine_formula>`_
	formula.

	The bearing between latitude, longitude: :math:`(\phi_1, \lambda_1)` and :math:`(\phi_2, \lambda_2)` is given by

	.. math::

		a = \sin^2(\frac{\phi_2 - \phi_1}{2}) + \cos(\phi_1) \cos(\phi_2) \sin^2(\frac{\lambda_2 - \lambda_1}{2})

		d = 2 R_0 \cdot \mathrm{atan2}(\sqrt{a}, \sqrt{1-a})

	The earth radius :math:`R_0` is taken to be :py:data:`py1090.helpers.EARTH_RADIUS`. The approximation of a spherical earth is made.

	Args:
		lat1 (float): :math:`\phi_1`, the latitude of the reference location
		lon1 (float): :math:`\lambda_1`, the longitude of the reference location
		lat2 (float): :math:`\phi_2`, the latitude of the target location
		lon2 (float): :math:`\lambda_2`, the longitude of the target location

	Returns:
		float: the distance in meters.

	"""
	lambda1 = math.radians(lon1)
	lambda2 = math.radians(lon2)

	phi1 = math.radians(lat1)
	phi2 = math.radians(lat2)

	dphi = phi2-phi1
	dlambda = lambda2-lambda1

	a = math.pow(math.sin(dphi/2), 2) + math.cos(phi1)*math.cos(phi2)*math.pow(math.sin(dlambda/2), 2)
	dsigma = 2*math.atan2(math.sqrt(a), math.sqrt(1-a))

	distance = EARTH_RADIUS * dsigma

	return distance

def bearing_between(lat1, lon1, lat2, lon2):
	r"""Calculates the bearing angle between two locations, in radians.

	The bearing between latitude, longitude: :math:`(\phi_1, \lambda_1)` and :math:`(\phi_2, \lambda_2)` is given by

	.. math::

		\mathrm{atan2}(\sin(\lambda_2 - \lambda_1) \cos(\phi_1), \cos(\phi_2) \sin(\phi_1) - \sin(\phi_2) \cos(\phi_2) \cos(\lambda_2 - \lambda_1))

	Args:
		lat1 (float): :math:`\phi_1`, the latitude of the reference location
		lon1 (float): :math:`\lambda_1`, the longitude of the reference location
		lat2 (float): :math:`\phi_2`, the latitude of the target location
		lon2 (float): :math:`\lambda_2`, the longitude of the target location

	Returns:
		float: the bearing angle in radians, between :math:`-\pi` and :math:`\pi`.

	"""
	lambda1 = math.radians(lon1)
	lambda2 = math.radians(lon2)

	phi1 = math.radians(lat1)
	phi2 = math.radians(lat2)

	dphi = phi2-phi1
	dlambda = lambda2-lambda1

	bearing = math.atan2(math.sin(dlambda)*math.cos(phi1), \
		math.cos(phi2)*math.sin(phi1)-math.sin(phi2)*math.cos(phi1)*math.cos(dlambda))

	return bearing


def knots_to_kmh(knots):
	"""Converts velocity in knots to velocity in km/h.

	1 knot is 1 nm/h (nautical mile per hour) and 1.852 km/h.

	Args:
		knots (float): velocity in knots

	Returns:
		float: velocity in km/h

	"""
	return 1.852*knots

def knots_to_mps(knots):
	"""Converts velocity in knots to velocity in m/s (meters per second).

	1 knot is 1 nm/h (nautical mile per hour), 1.852 km/h and about 6.67 m/s.

	Args:
		knots (float): velocity in knots

	Returns:
		float: velocity in m/s

	"""
	kmh = knots_to_kmh(knots)
	return 3.6*kmh
